#ifdef _CH_
#pragma package <opencv>
#endif

#ifndef _EiC
#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <opencv/cvaux.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>
#include <ctype.h>
#include <sys/stat.h>
#endif
#include <fstream>
#include <signal.h>
#include <sys/time.h>

// using std::cout;
// using std::endl;

// using namespace cv;
// using namespace std;


const char *VENTANA="highspeedTGXf Receiver";

CvCapture* capture = 0;
IplImage* frame;
IplImage* imagen_color;
CvMat imagen_data;


// The following CRC code comes from RFC1952, Section 8
//   http://tools.ietf.org/html/rfc1952#section-8

/* Table of CRCs of all 8-bit messages. */
unsigned long crc_table[256];

/* Flag: has the table been computed? Initially false. */
int crc_table_computed = 0;

/* Make the table for a fast CRC. */
void make_crc_table(void)
{
  unsigned long c;
  int n, k;
  for (n = 0; n < 256; n++) {
    c = (unsigned long) n;
    for (k = 0; k < 8; k++) {
      if (c & 1) {
        c = 0xedb88320L ^ (c >> 1);
      } else {
        c = c >> 1;
      }
    }
    crc_table[n] = c;
  }
  crc_table_computed = 1;
}


unsigned long update_crc(unsigned long crc,
  unsigned char *buf, int len)
{
  unsigned long c = crc ^ 0xffffffffL;
  int n;

  if (!crc_table_computed)
    make_crc_table();
  for (n = 0; n < len; n++) {
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  }
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
unsigned long crc(unsigned char *buf, int len)
{
  return update_crc(0L, buf, len);
}




/*

                                   _JNJ`
                               .JNMH`
                              JMMF`       `;.
                            .NMM)          `MN.
                            MMM)            (MML
                           (MMM`             MMML
                     M  I  D  N  I  G  H  T    C  o  D  E
                           (NMMF             MHNH
                            NMML            .MMM
                             NMML          .NMH
                              4MMNL       .#F
                               `4HNNL_   `
                                   `"""`

                          Copyright (C) 2004-2014
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/

#define TGXF_DEBUG 0
#define TGXF_BPP 1	// CHANGE TO MATCH THE BITS PER PIXEL SOURCE ENCODING
			// Valid options are 1 or 3 (24bpp is untested)
			// Yes, this should be a command line option

#define TGXF_CONTROL_TYPE_START 1
#define TGXF_CONTROL_TYPE_STOP 2
#define TGXF_CONTROL_TYPE_STATUS 3
#define TGXF_CONTROL_SUBTYPE_START_FILENAME 1
#define TGXF_CONTROL_SUBTYPE_START_FILESIZE 2
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION 3
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS 4
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES 5
#define TGXF_CONTROL_SUBTYPE_STOP_PAUSE 1
#define TGXF_CONTROL_SUBTYPE_STOP_COMPLETE 2
#define TGXF_CONTROL_SUBTYPE_STOP_CANCEL 3
#define TGXF_CONTROL_SUBTYPE_STATUS_SINCE 1


// TGXf Session Parameters (TGXf State Machine)
unsigned int tgxf_session_total_frames = 0;
unsigned int tgxf_session_first_data_frame = 0;
unsigned int tgxf_session_good_start = 0;
unsigned int tgxf_session_last_data_counter = 15;
unsigned int tgxf_session_frame_count = 0;
unsigned int tgxf_session_error_count = 0;
unsigned long tgxf_session_recent_crc32 = 0L;
unsigned long tgxf_session_total_crc32 = 0L;
unsigned int tgxf_session_max_errors = 5;
unsigned int tgxf_session_errors = 0;
char * tgxf_session_error_str = NULL;
unsigned int tgxf_session_error_size = 0;
unsigned int tgxf_session_error_len = 0;
unsigned int tgxf_session_aborted = 0;
unsigned long tgxf_session_total_bytes = 0; // CHANGED to 32bit for hsTGXf
unsigned int tgxf_session_complete = 0;
unsigned int tgxf_session_max_timeout = 5;
unsigned int tgxf_session_timeout_ticks = 0;
unsigned int tgxf_session_paused = 0;

unsigned int tgxf_session_debug_image_counter = 0;


// TGXf Session Parameters (TGXf Global Options)
unsigned char debug_session_file_name[100];
unsigned char tgxf_session_file_name[20];
char * tgxf_session_filepath;
unsigned long tgxf_session_file_size = 0;  // CHANGED to 32bit for hsTGXf
unsigned int tgxf_session_qrcode_version = 0;
unsigned int tgxf_session_qrcode_fps = 0;
unsigned long tgxf_session_qrcode_bytes = 0; // CHANGED to 32bit for hsTGXf



unsigned int get_size_for_file(char *filepath) {
  struct stat st;
  unsigned int size;

  stat(filepath, &st);
  size = st.st_size;

  return size;
}


unsigned long get_crc_for_file(char *filepath) {
  unsigned char data_buffer[1024];
  unsigned char *data_buffer_p;
  int data_buffer_size = 1024;
  unsigned int read_len;
  unsigned long crc;
  FILE *fp;

  crc = 0L;
  if(!filepath)
    return crc;

  if((fp = fopen((const char *)filepath, "rb")) == NULL)
    return crc;

  data_buffer_p = data_buffer;
  while((read_len = fread(data_buffer, 1, data_buffer_size, fp)) > 0) {
    crc = update_crc(crc, data_buffer_p, read_len);
  }

  return crc;
}


unsigned int unpack_control_payload_16bit_number(unsigned char *frame_payload, int payload_size)
{

  if(payload_size >= 2)
    return (unsigned int)frame_payload[0] << 8 | (unsigned int)frame_payload[1];

  return 0;
}


unsigned long unpack_control_payload_32bit_number(unsigned char *frame_payload, int payload_size)
{
  if(payload_size >= 4) {
    return (unsigned long)frame_payload[0] << 24 | (unsigned long)frame_payload[1] << 16 |
      (unsigned long)frame_payload[2] << 8 | (unsigned long)frame_payload[3]; 
  }

  return 0L;
}


unsigned int unpack_control_payload_clean_string(unsigned char *string_dest, unsigned char *frame_payload, int payload_size)
{
  unsigned char *chr_p;

  memset(string_dest, 0, 10);
  if(payload_size > 0) {
    // Store String
    if(memcpy(string_dest, frame_payload, 9) != string_dest) {
      fprintf(stderr, "ERROR(TGXf): badly formatted control string.\n");
      return -1;
    }
    string_dest[10] = '\0';
    // Clean String
    chr_p = string_dest;
    while(*chr_p) {
      // Unsavory characters are replaced
      if(*chr_p < 32 || ( *chr_p > 32 && *chr_p < 45 ) ||
        ( *chr_p > 46 && *chr_p < 48 ) || ( *chr_p > 57 && *chr_p < 65 ) ||
        ( *chr_p > 90 && *chr_p < 97 ) || *chr_p > 122) {
          *chr_p = '_';
      }
      // No leading dot
      if(*chr_p == 46 && chr_p == string_dest) {
          *chr_p = '_';
      }
      chr_p++;
    }
  }

  return 0;
}



void tgxf_timeout_watchdog(int sig) {
  if(tgxf_session_complete) {
    tgxf_session_timeout_ticks = 0;
    if(TGXF_DEBUG)
      printf("DEBUG(TGXf): Session complete ..\n");
    return;
  }
  if(tgxf_session_paused) {
    tgxf_session_timeout_ticks = 0;
    if(TGXF_DEBUG)
      printf("DEBUG(TGXf): Session paused ..\n");
    return;
  }
  if(tgxf_session_timeout_ticks >= tgxf_session_max_timeout) {
    if(!tgxf_session_aborted) {
      printf("(TGXf): Session receive timeout exceeded, aborting xfer.\n");
      tgxf_session_aborted = 1;
    }
  }
  tgxf_session_timeout_ticks++;

  return;
}


static int decode_tgxf_frame(unsigned char * packet_buffer, unsigned int packet_len) {
  struct itimerval tbuf;
  struct sigaction action;
  unsigned char *frame_buffer;
  unsigned char *frame_payload;
  char error_buffer[1024];
  int frame_size;
  int payload_size;
  int error_buffer_size = 1024;

  unsigned char control_byte;
  unsigned int control_bit;
  unsigned int control_type;
  unsigned int control_subtype;
  unsigned int data_counter;
  unsigned int expected_counter;
  unsigned int lost_frames;
  unsigned long transfer_bytes_remaining; // CHANGED to 32bit for hsTGXf
  unsigned int frames_remaining;
  unsigned int minutes_remaining;
  unsigned int seconds_remaining;
  unsigned int bytes_to_use;
  unsigned int lost_frame_idx;
  FILE *fp;
  int i;
  float progress_state;

  frame_size = packet_len;
  if(frame_size <= 0) {
    fprintf(stderr, "ERROR(TGXf): invalid frame size: %d\n", frame_size);
    return -1;
  }
  if((frame_buffer = (unsigned char *)malloc(frame_size)) == NULL) {
    fprintf(stderr, "ERROR(TGXf): unable to allocate frame buffer\n");
    return -1;
  }
  if(memcpy(frame_buffer, packet_buffer, frame_size) != frame_buffer) {
    fprintf(stderr, "ERROR(TGXf): failed to establish frame buffer\n");
    return -1;
  }

  // Update watchdog timer
  tgxf_session_timeout_ticks = 0;

  // Point to Payload
  frame_payload = frame_buffer;
  frame_payload++;
  payload_size = frame_size - 1;

  // Get Control Byte and Bit
  control_byte = *frame_buffer;
  control_bit = control_byte & 1;

  switch(control_bit) {

    // CONTROL FRAME
    case 1:
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Control Frame\n");
      control_type = (control_byte & 14) >> 1;
      // Only accept STOP if session has been aborted by client
      if(tgxf_session_aborted && control_type != TGXF_CONTROL_TYPE_STOP) {
        if(TGXF_DEBUG)
          fprintf(stdout, "DEBUG(TGXf): ignoring non-STOP control message in aborted transfer.\n");
        break;
      }
      control_subtype = (control_byte & 240) >> 4;
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Control Type=[%d], SubType=[%d]\n",control_type, control_subtype);
      switch(control_type) {
        case TGXF_CONTROL_TYPE_START:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type START\n");
          // Don't accept START if data frames have been received
          if(tgxf_session_first_data_frame) {
            if(TGXF_DEBUG)
              fprintf(stdout, "DEBUG(TGXf): ignoring late START control message in active transfer.\n");
          }
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_START_FILENAME:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): Control SubType FILENAME\n");
              unpack_control_payload_clean_string(tgxf_session_file_name, frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/FILENAME [%s].\n", tgxf_session_file_name);
              break;
            case TGXF_CONTROL_SUBTYPE_START_FILESIZE:
              tgxf_session_file_size = unpack_control_payload_32bit_number(frame_payload, payload_size);  // CHANGED to 32bit for hsTGXf
              // tgxf_session_file_size++; // WTF
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/FILESIZE [%lu].\n", tgxf_session_file_size); // CHANGED to 32bit for hsTGXf
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION:
              tgxf_session_qrcode_version = unpack_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_VERSION [%d].\n", tgxf_session_qrcode_version);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS:
              tgxf_session_qrcode_fps = unpack_control_payload_16bit_number(frame_payload, payload_size);
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_FPS [%d].\n", tgxf_session_qrcode_fps);
              break;
            case TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES:
              tgxf_session_qrcode_bytes = unpack_control_payload_32bit_number(frame_payload, payload_size);  // CHANGED to 32bit for hsTGXf
              // tgxf_session_qrcode_bytes++; // WTF
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): START/QRCODE_BYTES [%lu].\n", tgxf_session_qrcode_bytes);  // CHANGED to 32bit for hsTGXf
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              free(frame_buffer);
              return -1;
          }
          break;

        case TGXF_CONTROL_TYPE_STOP:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type STOP\n");
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_STOP_PAUSE:
              tgxf_session_paused = !tgxf_session_paused;
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): STOP/PAUSE.\n");
              break;
            case TGXF_CONTROL_SUBTYPE_STOP_COMPLETE:
              tgxf_session_total_crc32 = unpack_control_payload_32bit_number(frame_payload, payload_size);
              if(!tgxf_session_good_start) {
                if(TGXF_DEBUG)
                  fprintf(stdout, "WARNING(TGXf): failed to get required start data, ignoring out of state STOP.\n");
                break;
              }
              tgxf_session_recent_crc32 = get_crc_for_file((char *)tgxf_session_file_name);
              if(tgxf_session_total_crc32 != tgxf_session_recent_crc32) {
                fprintf(stdout, "ERROR(TGXf): CRC doesn't match, calculated %lu, received %lu.\n",
                  tgxf_session_recent_crc32, tgxf_session_total_crc32);
                tgxf_session_complete = 1; 				// CHANGED for hsTGXf
              } else {
                tgxf_session_complete = 1;
                fprintf(stdout, "(TGXf): CRC validated, transfer successful.\n");
              }
              // if(TGXF_DEBUG)
                // fprintf(stdout, "DEBUG(TGXf): STOP/COMPLETE [%u].\n", tgxf_session_total_crc32);
              // SESSION RESET
              break;
            case TGXF_CONTROL_SUBTYPE_STOP_CANCEL:
              if(TGXF_DEBUG)
                fprintf(stdout, "DEBUG(TGXf): STOP/CANCEL.\n");
              // SESSION RESET
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              free(frame_buffer);
              return -1;
          }
          break;

        case TGXF_CONTROL_TYPE_STATUS:
          if(TGXF_DEBUG)
            fprintf(stdout, "DEBUG(TGXf): Control Type STATUS\n");
          switch(control_subtype) {
            case TGXF_CONTROL_SUBTYPE_STATUS_SINCE:
              tgxf_session_recent_crc32 = unpack_control_payload_32bit_number(frame_payload, payload_size);
              // if(TGXF_DEBUG)
                // fprintf(stdout, "DEBUG(TGXf): STATUS/SINCE [%u].\n", tgxf_session_recent_crc32);
              break;
            default:
              fprintf(stdout, "WARNING(TGXf): reserved control sub-type used, frame ignored.\n");
              free(frame_buffer);
              return -1;
          }
          break;

        default:
          fprintf(stdout, "WARNING(TGXf): reserved control type used, frame ignored.\n");
          free(frame_buffer);
          return -1;
      }
      break;

    // DATA FRAME
    case 0:
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): Data Frame ------------------------------\n");

      // Ignore data frames if session aborted
      if(tgxf_session_aborted) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): session aborted, ignoring data frames.\n");
        break;
      }

      // If this is the first Data Frame then do we have the right START info to continue?
      if(!tgxf_session_first_data_frame) {
        tgxf_session_first_data_frame = 1;
        fprintf(stdout, "-= TGXf =-\n");
        fprintf(stdout, "  Filename:                 %s\n", tgxf_session_file_name);
        fprintf(stdout, "  File size:                %lu\n", tgxf_session_file_size); // CHANGED to 32bit for hsTGXf
        fprintf(stdout, "  QRcode Version:           %d\n", tgxf_session_qrcode_version);
        fprintf(stdout, "  QRcode Frames per Second: %d\n", tgxf_session_qrcode_fps);
        fprintf(stdout, "  QRcode Bytes per Frame:   %lu\n", tgxf_session_qrcode_bytes); // CHANGED to 32bit for hsTGXf
        fprintf(stdout, "\n");
        // Do we have sufficient START information to perform the data transfer?
        if(strlen((const char *)tgxf_session_file_name) > 0 &&
          tgxf_session_file_size > 0 &&
          tgxf_session_qrcode_fps > 0 &&
          tgxf_session_qrcode_bytes > 0) {

            // Establish session state
            tgxf_session_good_start = 1;

            // Setup session calculations
            tgxf_session_total_frames = tgxf_session_file_size / tgxf_session_qrcode_bytes;
            if(tgxf_session_total_frames * tgxf_session_qrcode_bytes < tgxf_session_file_size)
              tgxf_session_total_frames++;

            // Setup the TGXf watchdog timer
            action.sa_handler = tgxf_timeout_watchdog;
            sigemptyset(&action.sa_mask);
            action.sa_flags = 0;
            if(sigaction(SIGALRM, &action, NULL) < 0) {
              fprintf(stderr, "ERROR(TGXf): Watchdog timer setup failed, session won't timeout.\n");
            } else {
              tbuf.it_interval.tv_sec  = 1;
              tbuf.it_interval.tv_usec = 0;
              tbuf.it_value.tv_sec = 1;
              tbuf.it_value.tv_usec = 0;
              if(setitimer(ITIMER_REAL, &tbuf, NULL) == -1)
                fprintf(stderr, "ERROR(TGXf): Watchdog timer setup failed, session won't timeout.\n");
            }
        }
      }
      if(!tgxf_session_good_start) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): failed to get required start data, session aborted.\n");
        tgxf_session_aborted = 1;
        break;
      }

      // Duplicate Data Frame Detection - Ignore
      data_counter = (control_byte & 30) >> 1;
      if(data_counter == tgxf_session_last_data_counter) {
        if(TGXF_DEBUG)
          fprintf(stdout, "WARNING(TGXf): duplicate data frame, frame ignored.\n");
        break;
      }

      // Missing Data Frame Detection - Accumulate Errors
      expected_counter = tgxf_session_last_data_counter + 1;
      if(expected_counter == 16)
        expected_counter = 0;

      // Frame error detection				// CHANGE FOR hsTGXf
      unsigned long expected_frame_crc;
      unsigned long calculated_frame_crc;
      expected_frame_crc = unpack_control_payload_32bit_number(frame_payload, 4);
      frame_payload += 4;
      payload_size -= 4;
      calculated_frame_crc = update_crc(0, frame_payload, tgxf_session_qrcode_bytes);
      if(calculated_frame_crc == expected_frame_crc) {
        fprintf(stdout, "(TGXf): Frame %d CRC validated (%lu).\n",
          expected_counter + 1, calculated_frame_crc);
      } else {
        if(tgxf_session_total_frames - tgxf_session_frame_count != 1) {
          fprintf(stdout, "ERROR(TGXf): Frame %d CRC doesn't match, calculated %lu, received %lu.\n",
            expected_counter + 1, calculated_frame_crc, expected_frame_crc);
          break;
        } else {
          // Last frame (smaller than qrcode_bytes)
          transfer_bytes_remaining = tgxf_session_file_size - (tgxf_session_qrcode_bytes * tgxf_session_frame_count);
          if(transfer_bytes_remaining < tgxf_session_qrcode_bytes) {
            calculated_frame_crc = update_crc(0, frame_payload, transfer_bytes_remaining);
            if(calculated_frame_crc == expected_frame_crc) {
              fprintf(stdout, "(TGXf): Frame %d CRC validated (%lu).\n",
                expected_counter + 1, calculated_frame_crc);
            } else {
              fprintf(stdout, "ERROR(TGXf): Frame %d CRC doesn't match, calculated %lu, received %lu.\n",
                expected_counter + 1, calculated_frame_crc, expected_frame_crc);
              break;
            }
          }
        }
      }

      // Missing Data Frame Detection - Accumulate Errors
      if(TGXF_DEBUG)
        fprintf(stdout, "DEBUG(TGXf): data_counter=%02d expected_counter=%02d; frame %d.\n",
          data_counter, expected_counter, tgxf_session_frame_count);
      if(data_counter != expected_counter) {
        lost_frames = data_counter - expected_counter;
        if(expected_counter > data_counter) {
          lost_frames = data_counter + 16 - expected_counter;
        }
        // Error Accumulation
        tgxf_session_errors += lost_frames;
        if(tgxf_session_errors >= tgxf_session_max_errors) {
          tgxf_session_aborted = 1;
          fprintf(stdout, "ERROR(TGXf): maximum errors for session reached, transfer aborted.\n");
        }
        // Accumulate lost frame numbers to report to user
        if(!tgxf_session_error_str) {
          tgxf_session_error_size = 1024;
          if((tgxf_session_error_str = (char *)malloc(tgxf_session_error_size)) == NULL) {
            fprintf(stderr, "ERROR(TGXf): unable to allocate error buffer\n");
            free(frame_buffer);
            return -1;
          }
          memset(tgxf_session_error_str, 0, tgxf_session_error_size);
        }
        for(lost_frame_idx = 0; lost_frame_idx < lost_frames; lost_frame_idx++) {
          memset(error_buffer, 0, error_buffer_size);
          if(strlen(tgxf_session_error_str) > 0) {
            snprintf(error_buffer, error_buffer_size, ",%d",
              tgxf_session_frame_count + lost_frame_idx + 1);
          } else {
            snprintf(error_buffer, error_buffer_size, "%d",
              tgxf_session_frame_count + lost_frame_idx + 1);
          }
          if(strlen(error_buffer) > tgxf_session_error_size - strlen(tgxf_session_error_str) - 2) {
            tgxf_session_error_size += 1024;
            if((tgxf_session_error_str =
              (char *)realloc(tgxf_session_error_str, tgxf_session_error_size)) == NULL) {
              fprintf(stderr, "ERROR(TGXf): unable to allocate error buffer\n");
              free(frame_buffer);
              return -1;
            }
          }
          strcat(tgxf_session_error_str, error_buffer);
          if(tgxf_session_aborted) {
            strcat(tgxf_session_error_str, "+");
            break;
          }
        }
        if(lost_frames == 1) {
          fprintf(stdout, "WARNING(TGXf): lost 1 data frame, error recorded.\n");
        } else {
          fprintf(stdout, "WARNING(TGXf): lost %d data frames, error recorded.\n", lost_frames);
        }
        // Correct the session frame count 
        tgxf_session_frame_count += lost_frames;
      }

      // Do last-packet calculation and under-run detection
      transfer_bytes_remaining = 
        tgxf_session_file_size - (tgxf_session_qrcode_bytes * tgxf_session_frame_count);
      bytes_to_use = tgxf_session_qrcode_bytes;
      if(transfer_bytes_remaining < tgxf_session_qrcode_bytes)
        bytes_to_use = transfer_bytes_remaining;
      tgxf_session_total_bytes += bytes_to_use;
      if(TGXF_DEBUG) {
        fprintf(stdout, "DEBUG(TGXf): payload_size=%d bytes_to_use=%d qrcode_bytes=%lu; frame %d.\n", payload_size, bytes_to_use, tgxf_session_qrcode_bytes, tgxf_session_frame_count); // CHANGED to 32bit for hsTGXf
        fprintf(stdout, "DEBUG(TGXf): session_total_bytes after this frame is written = %lu/%lu.\n", tgxf_session_total_bytes, tgxf_session_file_size); // CHANGED to 32bit for hsTGXf
      }
      if(payload_size < bytes_to_use) {
        fprintf(stdout, "ERROR(TGXf): undersized payload on frame %d.\n", tgxf_session_frame_count);
        break;
      }

      // Update Display
      if(tgxf_session_errors) {
        if(tgxf_session_error_len < strlen(tgxf_session_error_str)) {
          tgxf_session_error_len = strlen(tgxf_session_error_str);
          fprintf(stdout, "\n");
          fprintf(stdout, "(TGXf): Frame errors: %s\n", tgxf_session_error_str);
        }
      }
      frames_remaining = tgxf_session_total_frames - tgxf_session_frame_count;
      seconds_remaining = frames_remaining / tgxf_session_qrcode_fps;
      minutes_remaining = seconds_remaining / 60;
      seconds_remaining = seconds_remaining % 60;
      progress_state = (float)20 - ((float)frames_remaining / (float)tgxf_session_total_frames * (float)20);
      fprintf(stdout, "(TGXf): Frame: %4d/%d ",
        tgxf_session_frame_count + 1, tgxf_session_total_frames);
      fprintf(stdout, " Errors: %2d/%d ",
        tgxf_session_errors, tgxf_session_max_errors);
      fprintf(stdout, " [");
      for(i = 0; i < 20; i++ ) {
        if(i >= progress_state) {
          fprintf(stdout, "%c", ' ' );
        } else {
          fprintf(stdout, "%c", '=' );
        }
      }
      fprintf(stdout, "] %02d min %02d sec.\r",
        minutes_remaining, seconds_remaining);
      // if(TGXF_DEBUG || frames_remaining == 1) // see also comment below @ line 1025
        fprintf(stdout, "\n");

      // Write Data to File
      if(access((const char *)tgxf_session_file_name, F_OK) == -1) {
        if((fp = fopen((const char *)tgxf_session_file_name, "w+b")) == NULL) {
          fprintf(stdout, "ERROR(TGXf): unable to create/open file for reading/writing, at %s.\n", tgxf_session_file_name);
          break;
        }
      } else {
        if((fp = fopen((const char *)tgxf_session_file_name, "r+b")) == NULL) {
          fprintf(stdout, "ERROR(TGXf): unable to open file for reading/writing, at %s.\n", tgxf_session_file_name);
          break;
        }
        if(fseek(fp, (tgxf_session_qrcode_bytes * tgxf_session_frame_count), SEEK_SET) < 0) {
          fprintf(stdout, "ERROR(TGXf): unable to seek file, at %s.\n", tgxf_session_file_name);
          fclose(fp);
          break;
        }
      }
      if(fwrite((const void *)frame_payload, (size_t)bytes_to_use, (size_t)1, fp) != (size_t)1) {
        fprintf(stdout, "ERROR(TGXf): unable to write to file, at %s.\n", tgxf_session_file_name);
        fclose(fp);
        break;
      }
      fclose(fp);

      // Increment tally
      tgxf_session_last_data_counter = data_counter;
      tgxf_session_frame_count++;
      break;

  }

  fflush(stdout);
  free(frame_buffer);
  return(1);
}




int get_tgxf_packet(unsigned char * buffer, unsigned int buf_len,
  int x_offset, int y_offset, int x_len, int y_len, int bpp) {
  int x, y;
  int r, g, b;
  int first_five_count;
  int first_five_blank;
  int pixel_type;
  unsigned char octet;
  int i;

  if(buf_len < (x_len * y_len * bpp / 8)) {
    printf("Receive buffer %d bytes, too small (%d byte packets).\n",
      buf_len, (x_len * y_len * bpp / 8));
    return -1;
  }
  if(bpp != 3 && bpp != 1) {
    printf("Bits-per-Pixel set to %d, not 3 or 1, skipping frame.\n",
      bpp);
    return -1;
  }

  first_five_blank = 0;
  first_five_count = 0;
  pixel_type = 0;
  i = 0;
  for(y = y_offset; y < (y_offset + y_len); y++) {
    // printf("\n");
    for(x = x_offset; x < (x_offset + x_len); x++) {
      // printf("DEBUG: x=%d, y=%d\n", x, y);
      r = CV_IMAGE_ELEM( imagen_color, uchar, y, x * 3 + 2 );
      g = CV_IMAGE_ELEM( imagen_color, uchar, y, x * 3 + 1 );
      b = CV_IMAGE_ELEM( imagen_color, uchar, y, x * 3 );
      // 3bpp
      if(bpp <= 3) {
        if(r < 128) {
          r = 0;
        } else {
          r = 1;
        }
        if(g < 128) {
          g = 0;
        } else {
          g = 1;
        }
        if(b < 128) {
          b = 0;
        } else {
          b = 1;
        }
        first_five_count++;
        if(r == 0 && g == 0 && b == 0) {
          first_five_blank++;
        }
        if(first_five_count == (5*8) && first_five_blank == (5*8)) {
          printf("Packet ignored per first five count/blank rule.\n");
          return -2;
        }
      }
      switch(bpp) {
        case 3:
          switch(pixel_type) {
            case 0:
              octet = 0;
              octet += (r << 7);
              octet += (g << 6);
              octet += (b << 5);
              pixel_type = 1;
              break;
            case 1:
              octet += (r << 4);
              octet += (g << 3);
              octet += (b << 2);
              pixel_type = 2;
              break;
            case 2:
              octet += (r << 1);
              octet += g;
              buffer[i] = octet;
              i++;
              // if(y == y_offset && x < x_offset + 50)
                // printf("%02x", octet);
              // printf("%02x", octet);
              octet = 0;
              octet += (b << 7);
              pixel_type = 3;
              break;
            case 3:
              octet += (r << 6);
              octet += (g << 5);
              octet += (b << 4);
              pixel_type = 4;
              break;
            case 4:
              octet += (r << 3);
              octet += (g << 2);
              octet += (b << 1);
              pixel_type = 5;
              break;
            case 5:
              octet += r;
              buffer[i] = octet;
              i++;
              // if(y == y_offset && x < x_offset + 50)
                // printf("%02x", octet);
              // printf("%02x", octet);
              octet = 0;
              octet += (g << 7);
              octet += (b << 6);
              pixel_type = 6;
              break;
            case 6:
              octet += (r << 5);
              octet += (g << 4);
              octet += (b << 3);
              pixel_type = 7;
              break;
            case 7:
              octet += (r << 2);
              octet += (g << 1);
              octet += b;
              buffer[i] = octet;
              // if(y == y_offset && x < x_offset + 50)
                // printf("%02x", octet);
              i++;
              // printf("%02x", octet);
              pixel_type = 0;
              break;
          }
          break;
        case 1:
          switch(pixel_type) {
            case 0:
              octet = 0;
              octet += (r << 7);
              // printf("[%d", r);
              pixel_type = 1;
              break;
            case 1:
              octet += (r << 6);
              // printf("%d", r);
              pixel_type = 2;
              break;
            case 2:
              octet += (r << 5);
              // printf("%d", r);
              pixel_type = 3;
              break;
            case 3:
              octet += (r << 4);
              // printf("%d", r);
              pixel_type = 4;
              break;
            case 4:
              octet += (r << 3);
              // printf("%d", r);
              pixel_type = 5;
              break;
            case 5:
              octet += (r << 2);
              // printf("%d", r);
              pixel_type = 6;
              break;
            case 6:
              octet += (r << 1);
              // printf("%d", r);
              pixel_type = 7;
              break;
            case 7:
              octet += r;
              // printf("%d]", r);
              buffer[i] = octet;
              i++;
              // printf("%02x ", octet);
              pixel_type = 0;
              break;
         }
      }
      // printf("row: %d, col: %d :: %d,%d,%d\n", y, x, r, g, b);
    }
  }

  return 0;
}


int check_pixel(IplImage* image, int x, int y, int r_cmp, int g_cmp, int b_cmp) {
  int r, g, b;
  int r_ck, g_ck, b_ck;

  if(y < 0 || x < 0)
    return -9;
  if(y > image->height)
    return -9;
  if(x > image->width)
    return -9;

  r_ck = 0;
  r = CV_IMAGE_ELEM(image, uchar, y, x * 3 + 2);
  if(r_cmp == 0 && r < 128)
    r_ck = 1;
  if(r_cmp == 1 && r > 127)
    r_ck = 1;

  g_ck = 0;
  g = CV_IMAGE_ELEM(image, uchar, y, x * 3 + 1);
  if(g_cmp == 0 && g < 128)
    g_ck = 1;
  if(g_cmp == 1 && g > 127)
    g_ck = 1;

  b_ck = 0;
  b = CV_IMAGE_ELEM(image, uchar, y, x * 3);
  if(b_cmp == 0 && b < 128)
    b_ck = 1;
  if(b_cmp == 1 && b > 127)
    b_ck = 1;

  if(r_ck && g_ck && b_ck) {
    // printf("check_pixel: %d,%d = %d(%d),%d(%d),%d(%d)\n", x, y, r_ck, r, g_ck, g, b_ck, b);
    return 1;
  }

  return 0;
}


int main( int argc, char** argv ) {
  CvScalar avg;
  CvScalar avgStd;
  CvScalar color;
  CvRect rect;
  CvPoint pt1, pt2;

  // TGXf Setup
  int have_bounding_rect = 0;
  int bounding_rect_gone = 0;
  int bounding_box_x = 0;
  int bounding_box_y = 0;
  int bounding_color = 0;
  int discovery_notice = 0;
  int bounding_ck = 0;
  int last_data_counter = -1;
  int expected_counter = 0;
  int lost_frames = 0;
  int data_counter = 0;
  char * output_buffer;

  unsigned char * packet_buffer;
  unsigned int packet_width = 1240;
  unsigned int packet_height = 650;
  // unsigned int packet_len = packet_width * packet_height * 1 / 8 + 1;
  unsigned int packet_len = packet_width * packet_height * TGXF_BPP / 8 + 1;

  srand ( time(NULL) );

  // Welcome
  printf("highspeedTGXf - Receiver - Proof-of-Concept by Ian Latter, MidnightCode.org\n"
    "\n"
    "  With much thanks to David Millan Escriva | Damiles\n"
    "\n");

  // Build Info
  // printf("BuildInfo %s\n", getBuildInformation());

  // Check if user specify image to process
  if(argc != 2 ) {
    printf("Use:\n\t%s video.mp4\n",argv[0]);
    return 0;
  }

  char* video_file = argv[1];

  printf("\n");
  // DEBUG - experiment for BlackMagic 4K
  // imagen_color = cvLoadImage(video_file);
  // cvNamedWindow(VENTANA, CV_WINDOW_NORMAL);
  // cvShowImage(VENTANA, imagen_color);
  // while(1) {
    // char c=cvWaitKey(30);
    // if (c==27) break;
  // }

  // Handle video input
  printf("Opening capture file ..\n");
  capture = cvCreateFileCapture(video_file);
  if(!capture) {
    printf("Failed to open video file: %s\n", video_file);
    return -1;
  }
  printf("Retrieving first frame from capture file ..\n");
  frame = cvQueryFrame(capture);
  // Increase color depth
  // --> http://opencv-srf.blogspot.com.au/2010/09/opencv-basics.html

  // packet buffer
  printf("Allocating packet buffer ..\n");
  if(packet_len <= 0) {
    fprintf(stderr, "Invalid packet buffer size: %d\n", packet_len);
    return -1;
  }
  if((packet_buffer = (unsigned char *)malloc(packet_len)) == NULL) {
    fprintf(stderr, "Unable to allocate packet buffer (%d bytes)\n", packet_len);
    return -1;
  }

  have_bounding_rect = 0;
  bounding_rect_gone = 0;

  // Create window
  printf("Creating display window ..\n");
  cvNamedWindow(VENTANA, CV_WINDOW_NORMAL);

  while(1) {
    frame = cvQueryFrame(capture);
    if(!frame)
      break;
    if(TGXF_DEBUG)
      printf("Frame depth %d\n", frame->depth);

    if(!imagen_color) {
      imagen_color = cvCreateImage(cvGetSize(frame), IPL_DEPTH_16U, 3);
      // imagen_color = cvCreateImage(cvGetSize(frame), frame->depth, 3);
      // imagen_color = cvCreateImage(cvGetSize(frame), frame->depth, 1);
      // imagen_color = cvCreateImage(cvGetSize(frame), IPL_DEPTH_8U, 3);
    }
    // cvCvtColor(frame, imagen_color, CV_RGB2RGB);
    // cvCopy(frame, imagen_color, 0);
    imagen_color=cvCloneImage(frame);

    // if(TGXF_DEBUG)  // see also comment above @ line 674
    fprintf(stdout, "-------------------------- [frame] --------------------------\n");
    fflush(stdout);


    // get the image data
    // height    = img->height;
    // width     = img->width;
    // step      = img->widthStep;
    // channels  = img->nChannels;
    // data      = (uchar *)img->imageData;
    // printf("Processing a %dx%d image with %d channels\n",
    //   imagen_color->height,imagen_color->width,imagen_color->nChannels); 
    int i,j,k;
    int min_val, max_val;
    int found_contiguous_column_offset;
    int found_contiguous_column_count;
    int found_contiguous_row_offset;
    int found_contiguous_row_count;
    uchar b, g, r;

    min_val = 0;
    max_val = 0;
    if(!have_bounding_rect) {
      bounding_box_x = 0;
      bounding_box_y = 0;
      found_contiguous_row_offset = 0;
      found_contiguous_row_count = 0;
      for(i=0; i<imagen_color->height; i++) {
        found_contiguous_column_offset = 0;
        found_contiguous_column_count = 0;
        for(j=0; j<imagen_color->width; j++) {
          r = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 + 2 );
          g = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 + 1 );
          b = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 );
          if(r > 128 && b < 128 && g < 128) {
            // fresh find
            if(found_contiguous_column_offset == 0) {
              found_contiguous_column_offset = j;
              found_contiguous_column_count = 1;
            } else {
              // contiguous chunk
              if(found_contiguous_column_offset + found_contiguous_column_count == j) {
                found_contiguous_column_count++;
              } else {
                // cautiously start again
                if(found_contiguous_column_count < packet_width) {
                  found_contiguous_column_offset = j;
                  found_contiguous_column_count = 1;
                }
              }
            }
          }
        }
        if(found_contiguous_column_count == packet_width) {
          if(bounding_box_x == 0 && bounding_box_y == 0) {
            bounding_box_x = found_contiguous_column_offset;
            bounding_box_y = i;
            if(TGXF_DEBUG)
              printf("b_x: %d, b_y: %d\n", bounding_box_x, bounding_box_y);
          }
          if(TGXF_DEBUG)
            printf("row: %d, col: %d, count: %d\n",
              i, found_contiguous_column_offset, found_contiguous_column_count);
          // fresh find
          if(found_contiguous_row_offset == 0) {
            found_contiguous_row_count = 1;
            found_contiguous_row_offset = i;
          } else {
            // contiguous chunk
            if(found_contiguous_row_offset + found_contiguous_row_count == i) {
              found_contiguous_row_count++;
            } else {
              // cautiously start again
              if(found_contiguous_row_count < packet_height) {
                found_contiguous_row_offset = i;
                found_contiguous_row_count = 1;
              }
            }
          }
        }
      }
      if(found_contiguous_row_count > (packet_height - 2) &&
        bounding_box_x + packet_width <= imagen_color->width &&
        bounding_box_y + packet_height <= imagen_color->height) {
      // if(found_contiguous_row_count == 400) {
        have_bounding_rect = 1;
        // manual correction !?
        // bounding_box_y--; 
        // found_contiguous_row_count++;
        // found_contiguous_row_count = 400;
        // bounding_box_y++;
        if(TGXF_DEBUG)
          printf("FOUND row: %d, col: %d -> count(%d)\n",
            bounding_box_x, bounding_box_y, found_contiguous_row_count);

        // Debug First row
        for(i=bounding_box_y; i < (2 + bounding_box_y); i++) {
          for(j=bounding_box_x; j< (10 + bounding_box_x); j++) {
            r = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 + 2 );
            g = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 + 1 );
            b = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 );
            // printf("row: %d, col: %d :: %d,%d,%d\n", i, j, r, g, b);
          }
        }
      }
    }
    if(have_bounding_rect && !bounding_rect_gone &&
      bounding_box_x > 0 && bounding_box_y > 0) {

      if(!discovery_notice) {
        printf("Have bounding (%d,%d)\n", bounding_box_x, bounding_box_y);
        discovery_notice++;
      }

      bounding_ck = 0;
      bounding_ck += check_pixel(imagen_color,
        bounding_box_x + 1, bounding_box_y + 1,
        1, 0, 0);
      bounding_ck += check_pixel(imagen_color,
        bounding_box_x + packet_width - 1, bounding_box_y + 1,
        1, 0, 0);
      bounding_ck += check_pixel(imagen_color,
        bounding_box_x + packet_width - 1,
        bounding_box_y + packet_height - 1,
        1, 0, 0);
      bounding_ck += check_pixel(imagen_color,
        bounding_box_x + 1,
        bounding_box_y + packet_height - 1,
        1, 0, 0);
      if(bounding_ck == 0) {
        printf("Bounding sync gone, now processing TGXf packets in bounding area.\n");
        bounding_rect_gone = 1;
      }
    }
    if(have_bounding_rect && bounding_rect_gone &&
      bounding_box_x > 0 && bounding_box_y > 0) {

      // Process bounded area
      if(get_tgxf_packet(packet_buffer, packet_len,
        bounding_box_x, bounding_box_y, packet_width, packet_height, TGXF_BPP) == 0) {
        // printf("Packet [%s]\n", packet_buffer);
        if(TGXF_DEBUG)
          printf("Have packet ... decoding ..\n");
        decode_tgxf_frame(packet_buffer, packet_len);
      }

      // Debug first row
      for(i=bounding_box_y; i < (2 + bounding_box_y); i++) {
        for(j=bounding_box_x; j< (10 + bounding_box_x); j++) {
          r = ((uchar *)(imagen_color->imageData + imagen_color->widthStep*i))[j*3+1];
          // r = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 + 2 );
          g = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 + 1 );
          b = CV_IMAGE_ELEM( imagen_color, uchar, i, j * 3 );
          // if(TGXF_DEBUG)
            // printf("xrow: %d, col: %d :: %d,%d,%d\n", i, j, r, g, b);
        }
      }

      // Paint bounding box
      bounding_color+=4;
      if(bounding_color > 255)
        bounding_color = 0;
      color = CV_RGB( bounding_color, 0, 0 );
      pt1.x = bounding_box_x;
      pt2.x = bounding_box_x + packet_width;
      pt1.y = bounding_box_y;
      pt2.y = bounding_box_y + packet_height;
      cvRectangle(imagen_color, pt1, pt2, color, 1, 8, 0); 
    }

    // http://stackoverflow.com/questions/7535364/how-do-i-access-a-pixel-in-opencv
    // http://note.sonots.com/OpenCV/MatrixOperations.html

    cvShowImage(VENTANA, imagen_color);

    char c=cvWaitKey(30);
    if (c==27) break;
  }

  cvReleaseCapture( &capture );
  cvDestroyWindow(VENTANA);

  return 0;
}

