#!/bin/bash

###########################################################
#
# OpenCV 3.0.0 alpha - instalao
# http://opencv.org/
#
###########################################################

#----------------------------------------------------------
# Criando um logger para registrar a instalao
#----------------------------------------------------------
# Tempo: incio
dateformat="+%a %b %-eth %Y %I:%M:%S %p %Z"
starttime=$(date "$dateformat")
starttimesec=$(date +%s)

# Pega o diretrio atual
curdir=$(cd `dirname $0` && pwd)

# Cria o arquivo onde as aes ficaro registradas
logfile="$curdir/install-opencv.log"
rm -f $logfile

# Logger simples
log(){
	timestamp=$(date +"%Y-%m-%d %k:%M:%S")
	echo "\n$timestamp $1"
	echo "$timestamp $1" >> $logfile 2>&1
}
 
# Iniciando a instalao do OpenCV 3.0.0
log "Iniciando a instalao do OpenCV 3.0.0"


#----------------------------------------------------------
# Assegurando um ambiente atualizado
#----------------------------------------------------------

# Informa ao usurio a prxima ao
log "Executando apt-get update e apt-get upgrade"
 
# Executa a ao
sudo apt-get update
sudo apt-get upgrade

 
#----------------------------------------------------------
# Instalando os pacotes das dependncias
#----------------------------------------------------------

log "Instalando as dependncias"
 
# Executa a ao
sudo apt-get -y install libopencv-dev build-essential cmake git libgtk2.0-dev pkg-config python-dev python-numpy libdc1394-22 libdc1394-22-dev libjpeg-dev libpng12-dev libtiff4-dev libjasper-dev libavcodec-dev libavformat-dev libswscale-dev libxine-dev libgstreamer0.10-dev libgstreamer-plugins-base0.10-dev libv4l-dev libtbb-dev libqt4-dev libfaac-dev libmp3lame-dev libopencore-amrnb-dev libopencore-amrwb-dev libtheora-dev libvorbis-dev libxvidcore-dev x264 v4l-utils unzip
 
 
#----------------------------------------------------------
# Instalando OpenCV
#----------------------------------------------------------

log "Baixando a biblioteca OpenCV 3.0.0"
 
# Definio de constante
FOLDER_NAME="opencv"
 
# Cria um novo diretrio para armazenar o cdigo-fonte
mkdir ${FOLDER_NAME}
 
# Entra no diretrio
cd ${FOLDER_NAME}
 
# Baixa o cdigo-fonte
# wget https://github.com/Itseez/opencv/archive/3.0.0-alpha.zip -O opencv-3.0.0-alpha.zip
cp ../opencv-3.0.0.zip .
 
# Extrai o contedo
unzip opencv-3.0.0.zip

log "Instalando a biblioteca OpenCV 3.0.0"
 
# Entra no diretrio
cd opencv-3.0.0
 
# Cria um diretrio chamado 'build'
mkdir build
 
# Entra no diretrio
cd build
 
# Constri o projeto usando CMake
cmake -D CMAKE_BUILD_TYPE=RELEASE -D CMAKE_INSTALL_PREFIX=/usr/local -D WITH_TBB=ON -D BUILD_NEW_PYTHON_SUPPORT=ON -D WITH_V4L=ON -D WITH_QT=ON -D WITH_OPENGL=ON ..
 
# Compila o projeto
make -j $(nproc)
 
# Instala as bibliotecas na pasta apropriada
sudo make install
 
# Adiciona o caminho das bibliotecas do OpenCV aos caminhos de pesquisa de biblioteca padro do Ubuntu
sudo /bin/bash -c 'echo "/usr/local/lib" > /etc/ld.so.conf.d/opencv.conf'
 
# Atualiza os caminhos de pesquisa de biblioteca padro do Ubuntu
sudo ldconfig
 
# Informa ao usurio que o OpenCV foi instalado com sucesso!
log "OpenCV 3.0.0 foi instalado com sucesso!"

#----------------------------------------------------------
# Mostra tempo gasto com a instalao
#----------------------------------------------------------

# Tempo: fim
endtime=$(date "$dateformat")
endtimesec=$(date +%s)

# Mostra tempo gasto com a instalao
elapsedtimesec=$(expr $endtimesec - $starttimesec)
ds=$((elapsedtimesec % 60))
dm=$(((elapsedtimesec / 60) % 60))
dh=$((elapsedtimesec / 3600))
displaytime=$(printf "%02d:%02d:%02d" $dh $dm $ds)
log "Tempo gasto: $displaytime\n"
